# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.
import time
from asyncio import CancelledError, shield

import numpy as np

from karabo.common.scenemodel.api import (
    BoxLayoutModel, CheckBoxModel, DisplayCommandModel, DisplayLabelModel,
    DisplayListModel, DisplayStateColorModel, DoubleLineEditModel,
    EditableListModel, EditableRegexModel, IntLineEditModel, LabelModel,
    LineEditModel, LineModel, NDArrayGraphModel, SceneModel, TableElementModel,
    VectorGraphModel, write_scene)
from karabo.middlelayer import (
    AccessLevel, AccessMode, AlarmCondition, Bool, Configurable, DaqDataType,
    Device, Double, Float, Hash, InputChannel, Int32, Int64, KaraboError,
    MetricPrefix, NDArray, Node, OutputChannel, Overwrite, RegexString, Slot,
    State, String, UInt32, UInt64, Unit, VectorBool, VectorChar, VectorDouble,
    VectorFloat, VectorHash, VectorInt32, VectorInt64, VectorRegexString,
    VectorString, VectorUInt32, VectorUInt64, background, sleep, slot)

VECTOR_MAX_SIZE = 10

NDARRAY_SHAPE = (100, 200)

DEFAULT_OPT = ["a", "b", "c"]


class DataNode(Configurable):
    daqDataType = DaqDataType.TRAIN

    int32 = Int32(accessMode=AccessMode.READONLY)
    string = String(accessMode=AccessMode.READONLY)
    vecInt64 = VectorInt64(
        accessMode=AccessMode.READONLY,
        minSize=1,
        maxSize=VECTOR_MAX_SIZE)
    ndarray = NDArray(
        accessMode=AccessMode.READONLY,
        dtype=Float,
        shape=NDARRAY_SHAPE)


class ChannelNode(Configurable):
    node = Node(DataNode)


class VectorNode(Configurable):
    boolProperty = VectorBool(
        displayedName="Boolean properties",
        description="Vector boolean values",
        defaultValue=[True, False, True, False, True, False],
        minSize=1,
        maxSize=VECTOR_MAX_SIZE)

    charProperty = VectorChar(
        displayedName="Char Vector",
        description="Vector char values",
        minSize=1,
        maxSize=VECTOR_MAX_SIZE,
        defaultValue=b"ABCDEF")

    int32Property = VectorInt32(
        displayedName="Int32 Vector (No Limits)",
        description="Vector int32 values",
        defaultValue=[-1, 2, 3])

    uint32Property = VectorUInt32(
        displayedName="UInt32 Vector",
        description="Vector uint32 values",
        defaultValue=[1, 2, 3],
        minSize=1,
        maxSize=VECTOR_MAX_SIZE)

    int64Property = VectorInt64(
        displayedName="Int64 Vector (No Limits)",
        description="Vector int64 values",
        defaultValue=[1, 2],
        minSize=1,
        maxSize=VECTOR_MAX_SIZE)

    uint64Property = VectorUInt64(
        displayedName="UInt64 Vector",
        description="Vector Uint64 values",
        defaultValue=[10, 10],
        minSize=1,
        maxSize=VECTOR_MAX_SIZE)

    floatProperty = VectorFloat(
        displayedName="Float Vector (No Limits)",
        description="Vector float values",
        defaultValue=[7.1, 0.1, 5.3])

    doubleProperty = VectorDouble(
        displayedName="Double Vector",
        description="Vector double values",
        defaultValue=[1.23, 0.4, 1.0],
        minSize=1,
        maxSize=VECTOR_MAX_SIZE)

    stringProperty = VectorString(
        displayedName="String Vector",
        description="Vector string values",
        minSize=1,
        maxSize=VECTOR_MAX_SIZE,
        defaultValue=["A", "B", "C"])


class TableRow(Configurable):
    intProperty = UInt32(
        displayedName="Integer Property",
        description="An optional integer property",
        options=[1, 2, 3, 4],
        defaultValue=2)

    stringProperty = String(
        displayedName="String Property",
        description="An option property for strings",
        options=["spam", "eggs"],
        defaultValue="spam")

    bool = Bool(
        displayedName="Bool",
        defaultValue=False,
        accessMode=AccessMode.RECONFIGURABLE)

    boolReadOnly = Bool(
        displayedName="Bool (RO)",
        defaultValue=False,
        accessMode=AccessMode.READONLY)

    floatProperty = Float(
        displayedName="Float Property",
        description="A float property",
        defaultValue=-3.0)

    doubleProperty = Double(
        displayedName="Double Property",
        description="A double property",
        defaultValue=14.0)


class CounterNode(Configurable):

    @Slot(displayedName="Increment",
          description="Increment 'Counter read-only'")
    def increment(self):
        self.counterReadOnly = self.counterReadOnly + 1

    @Slot(displayedName="Reset",
          description="Reset 'Counter read-only'")
    def reset(self):
        self.counterReadOnly = 0

    @UInt32(displayedName="Counter",
            description="Values will be transferred to 'Counter read-only'",
            defaultValue=0)
    def counter(self, newValue):
        self.counter = newValue
        self.counterReadOnly = newValue

    counterReadOnly = UInt32(
        displayedName="Counter read-only",
        accessMode=AccessMode.READONLY,
        defaultValue=0)


class PropertyTest(Device):

    allowedStates = [
        State.INIT, State.STARTED, State.NORMAL, State.STARTING,
        State.STOPPING]

    state = Overwrite(
        defaultValue=State.INIT,
        options=allowedStates)

    @Bool(displayedName="Bool",
          description="a boolean value",
          defaultValue=False)
    def boolProperty(self, newValue):
        self.boolProperty = newValue
        self.boolPropertyReadOnly = newValue

    boolPropertyReadOnly = Bool(
        displayedName="Bool Readonly",
        description="A readonly boolean property",
        defaultValue=True,
        accessMode=AccessMode.READONLY)

    @Int32(displayedName="Int32",
           description="An integer property (32 Bit)",
           defaultValue=32_000_000)
    def int32Property(self, newValue):
        self.int32Property = newValue
        self.int32PropertyReadOnly = newValue

    int32PropertyReadOnly = Int32(
        displayedName="Int32 (RO)",
        description="An integer property",
        defaultValue=32_000_000,
        accessMode=AccessMode.READONLY)

    @UInt32(displayedName="UInt32",
            description="An unsigned integer property (32 Bit)",
            defaultValue=32_000_000)
    def uint32Property(self, newValue):
        self.uint32Property = newValue
        self.uint32PropertyReadOnly = newValue

    uint32PropertyReadOnly = UInt32(
        displayedName="UInt32 (RO)",
        description="A readonly integer property",
        defaultValue=32_000_000,
        accessMode=AccessMode.READONLY)

    @Int64(displayedName="Int64",
           description="An integer 64 Bit property",
           defaultValue=3_200_000_000)
    def int64Property(self, newValue):
        self.int64Property = newValue
        self.int64PropertyReadOnly = newValue

    int64PropertyReadOnly = Int64(
        displayedName="Int64 (RO)",
        description="A readonly integer property",
        defaultValue=3_200_000_000,
        accessMode=AccessMode.READONLY)

    @UInt64(displayedName="UInt64",
            description="An unsigned integer property (64 Bit)",
            defaultValue=3_200_000_000)
    def uint64Property(self, newValue):
        self.uint64Property = newValue
        self.uint64PropertyReadOnly = newValue

    uint64PropertyReadOnly = UInt64(
        displayedName="UInt64 (RO)",
        description="A readonly integer property (64 Bit)",
        defaultValue=3_200_000_000,
        accessMode=AccessMode.READONLY)

    @Float(displayedName="Float (Min / Max)",
           description="A float property",
           defaultValue=20.0,
           minExc=-1000.0,
           maxExc=1000.0)
    def floatProperty(self, newValue):
        self.floatProperty = newValue
        self.floatPropertyReadOnly = newValue

    floatPropertyReadOnly = Float(
        displayedName="Float (RO)",
        description="A readonly float property",
        defaultValue=0.0,
        accessMode=AccessMode.READONLY)

    @Double(displayedName="Double",
            description="A double property",
            defaultValue=0.1)
    def doubleProperty(self, newValue):
        self.doubleProperty = newValue
        self.doublePropertyReadOnly = newValue

    doublePropertyReadOnly = Double(
        displayedName="Double (RO)",
        description="A readonly double property for testing alarms",
        defaultValue=3.1415,
        accessMode=AccessMode.READONLY)

    stringProperty = String(
        displayedName="String",
        description="A string property",
        defaultValue="XFEL")

    regexProperty = RegexString(
        displayedName="Regex String",
        regex=r"^[A-Za-z0-9_\-]{1,20}$",
        description="A regex string property",
        defaultValue="RegexKarabo")

    vectorRegexProperty = VectorRegexString(
        displayedName="Vector Regex String",
        regex=r"^[A-Za-z0-9_\-]{1,20}$",
        description="A vector regex string property",
        defaultValue=["VectorRegexKarabo"])

    @Slot(displayedName="Set Alarm",
          description="Set alarm to value of String property - if convertable")
    def setAlarm(self):
        level = AlarmCondition(self.stringProperty)
        self.globalAlarmCondition = level

    vectors = Node(
        VectorNode,
        displayedName="Vectors",
        description="A node containing vector properties")

    table = VectorHash(
        TableRow,
        displayedName="Table property",
        description="Table containing one node",
        accessMode=AccessMode.RECONFIGURABLE,
        defaultValue=[
            Hash('intProperty', 1, 'stringProperty', 'spam',
                 'bool', False, 'boolReadOnly', False, 'floatProperty', 5.0,
                 'doubleProperty', 27.1),
            Hash('intProperty', 3, 'stringProperty', 'eggs',
                 'bool', True, 'boolReadOnly', True, 'floatProperty', 1.3,
                 'doubleProperty', 22.5)])

    # TODO: Add tableReadOnly = VectorHash(...)
    #       and make table a function changing it

    @InputChannel(displayedName="Input", raw=False)
    async def input(self, data, meta):
        procTimeSecs = self.processingTime.value / 1000.
        await sleep(procTimeSecs)

        self.inputCounter = self.inputCounter.value + 1
        self.currentInputId = data.node.int32

        await self._send_data_action()

    @input.endOfStream
    async def input(self, channel):
        self.inputCounterAtEos = self.inputCounter.value
        await self.output.writeEndOfStream()

    @input.close
    async def input(self, channel):
        self.logger.info(f"Close handler called by {channel}")

    processingTime = UInt32(
        displayedName="Processing Time",
        description="Processing time of input channel data handler",
        defaultValue=0,
        unitSymbol=Unit.SECOND,
        metricPrefixSymbol=MetricPrefix.MILLI,
        accessMode=AccessMode.RECONFIGURABLE,
        requiredAccessLevel=AccessLevel.EXPERT,
    )
    output = OutputChannel(
        ChannelNode,
        displayedName="Output")

    outputFrequency = Float(
        displayedName="Output frequency",
        description="The target frequency for continously writing to 'Output'",
        unitSymbol=Unit.HERTZ,
        accessMode=AccessMode.RECONFIGURABLE,
        requiredAccessLevel=AccessLevel.EXPERT,
        maxInc=1000.,
        minExc=0.0,
        defaultValue=1.)

    outputCounter = Int32(
        displayedName="Output counter",
        description="Last value sent as 'int32' via output channel 'Output'",
        defaultValue=0,
        accessMode=AccessMode.READONLY)

    currentInputId = Int32(
        displayedName="Current Input Id",
        description="Last value received as 'int32' on input channel",
        defaultValue=0,
        accessMode=AccessMode.READONLY)

    inputCounter = Int32(
        displayedName="Input counter",
        description="Value of 'Input Counter' when endOfStream was received",
        defaultValue=0,
        accessMode=AccessMode.READONLY)

    inputCounterAtEos = Int32(
        displayedName="Input counter @ EOS",
        defaultValue=0,
        accessMode=AccessMode.READONLY)

    availableScenes = VectorString(
        displayedName="Available Scenes",
        displayType="Scenes",
        description="Provides a scene for the Configuration Manager.",
        accessMode=AccessMode.READONLY,
        defaultValue=['scene'])

    async def onInitialization(self):
        self.state = State.NORMAL
        self.acquiring_task = None

    @Slot(displayedName="Start Writing",
          description="Start writing continously to output channel "
                      "'Output'",
          allowedStates=[State.NORMAL])
    async def startWritingOutput(self):
        self.state = State.STARTING
        self.acquiring_task = background(self._send_data_task)

    @Slot(displayedName="Stop Writing",
          description="Stop writing continously to output channel "
                      "'Output'",
          allowedStates=[State.STARTED])
    async def stopWritingOutput(self):
        self.state = State.STOPPING
        if self.acquiring_task:
            self.acquiring_task.cancel()
        else:
            self.state = State.NORMAL

    @Slot(displayedName="EOS to Output",
          description="Write end-of-stream to output channel 'Output'",
          allowedStates=[State.NORMAL])
    async def eosOutput(self):
        await self.output.writeEndOfStream()

    @Slot(displayedName="Write to Output",
          description="Write once to output channel 'Output'",
          allowedStates=[State.NORMAL])
    async def writeOutput(self):
        await self._send_data_action()

    @Slot(displayedName="Reset Channels",
          description="Reset counters involved in input/output channel "
                      "data flow",
          allowedStates=[State.NORMAL])
    async def resetChannelCounters(self):
        self.inputCounter = 0
        self.inputCounterAtEos = 0
        self.outputCounter = 0
        self.currentInputId = 0

    async def _send_data_action(self):
        outputCounter = self.outputCounter.value + 1
        output = self.output.schema.node
        output.int32 = outputCounter
        output.string = f'{outputCounter}'
        output.vecInt64 = [outputCounter] * VECTOR_MAX_SIZE
        output.ndarray = np.full(NDARRAY_SHAPE,
                                 outputCounter, dtype=np.float32)

        # XXX: implement image data

        async def write_and_count():
            await self.output.writeData()
            self.outputCounter = outputCounter

        # Prohibit task cancelation in between writing to output and
        # counter increase to keep them in sync.
        await shield(write_and_count())

    async def _send_data_task(self):
        self.state = State.STARTED
        while True:
            try:
                before = time.time()
                await self._send_data_action()
                delay = 1 / self.outputFrequency.value
                # Adjust the delay by what it took to send data
                delay = max(0., delay - (time.time() - before))
                # Cast since with numpy 2.0.1 delay is np.float32 and that
                # causes extra delay (with numpy 2.0.1).
                await sleep(float(delay))
            except CancelledError:
                self.state = State.NORMAL
                self.acquiring_task = None
                return

    node = Node(
        CounterNode,
        displayedName="Node")

    _available_macros = ['default', 'another', 'the_third']

    availableMacros = VectorString(
        displayedName="Available Macros",
        description="Provides macros from the device",
        accessMode=AccessMode.READONLY,
        defaultValue=_available_macros)

    def generate_macro(self, name):
        if name in self._available_macros[:2]:
            code = "from karabo.middlelayer import Macro, Slot, String\n\n" \
                   "class helloWorld(Macro):\n" \
                   f"    name = String(defaultValue='{name}')\n" \
                   "    @Slot()\n" \
                   "    def sayHello(self):\n" \
                   "        print(f'Hello, {self.name} macro!')\n"
        else:
            code = "import this\n"
        return code

    @slot
    def requestScene(self, params):
        """Fulfill a scene request from another device.

        :param params: A `Hash` containing the method parameters
        """
        payload = Hash('success', False)
        name = params.get('name', default='scene')
        if name == 'scene':
            payload.set('success', True)
            payload.set('name', name)
            payload.set('data', get_scene(self.deviceId))

        return Hash('type', 'deviceScene',
                    'origin', self.deviceId,
                    'payload', payload)

    @slot
    def requestMacro(self, params):
        payload = Hash('success', True)

        name = params.get('name', default='')
        payload.set('data', self.generate_macro(name))
        return Hash('type', 'deviceMacro', 'origin', self.deviceId, 'payload',
                    payload)

    faultyError = String(
        displayedName="Faulty Error",
        description="Configure the type of exception to be thrown",
        defaultValue="KaraboError",
        options=["KaraboError", "RuntimeError"]
    )

    @String(displayedName="Faulty String",
            description="A string property that could not be set from "
                        "the system",
            defaultValue="Karabo")
    def faultyString(self, value):
        if value != "Karabo":
            if self.faultyError == "RuntimeError":
                exception = RuntimeError
            else:
                exception = KaraboError
            raise exception(f"Only 'Karabo' is allowed here not '{value}'")
        else:
            self.faultyString = value

    @Slot(displayedName="Faulty Slot",
          description="A slot that throws an error",
          defaultValue="Karabo")
    def faultySlot(self):
        if self.faultyError == "RuntimeError":
            exception = RuntimeError
        else:
            exception = KaraboError
        raise exception("Faulty Slot cannot be executed")

    @VectorString(displayedName="Options for optString",
                  minSize=1,
                  defaultValue=DEFAULT_OPT)
    async def optionsForOptString(self, value):
        self.optionsForOptString = value
        background(self._injectStringOpt())

    async def _injectStringOpt(self):
        self.__class__.optString = Overwrite(
            options=self.optionsForOptString.value,
            defaultValue=self.optionsForOptString.value[0]
        )
        await self.publishInjectedParameters()

    optString = String(displayedName="String with Options",
                       defaultValue=DEFAULT_OPT[0],
                       options=DEFAULT_OPT)


def get_scene(deviceId):
    scene0 = DisplayStateColorModel(
        height=21.0, keys=[f'{deviceId}.state'],
        parent_component='DisplayComponent', show_string=True, width=231.0,
        x=70.0, y=40.0)
    scene10 = LabelModel(
        font='Source Sans Pro,10,-1,5,50,0,0,0,0,0', foreground='#000000',
        height=21.0, parent_component='DisplayComponent', text='Host',
        width=35.0, x=310.0, y=40.0)
    scene11 = DisplayLabelModel(
        font_size=10, height=21.0, keys=[f'{deviceId}.hostName'],
        parent_component='DisplayComponent', width=166.0, x=345.0, y=40.0)
    scene1 = BoxLayoutModel(
        height=21.0, width=201.0, x=310.0, y=40.0, children=[scene10, scene11])
    scene2 = LabelModel(
        font='Source Sans Pro,10,-1,5,50,0,0,0,0,0', foreground='#000000',
        height=21.0, parent_component='DisplayComponent', text='DeviceID',
        width=57.0, x=10.0, y=10.0)
    scene3 = DisplayLabelModel(
        font_size=10, height=21.0, keys=[f'{deviceId}.deviceId'],
        parent_component='DisplayComponent', width=441.0, x=70.0, y=10.0)
    scene400 = LabelModel(
        font='Source Sans Pro,10,-1,5,50,0,0,0,0,0', foreground='#000000',
        height=40.0, parent_component='DisplayComponent', text='Bool Readonly',
        width=88.0, x=20.0, y=90.0)
    scene401 = LabelModel(
        font='Source Sans Pro,10,-1,5,50,0,0,0,0,0', foreground='#000000',
        height=40.0, parent_component='DisplayComponent', text='Int32 (RO)',
        width=88.0, x=20.0, y=130.0)
    scene402 = LabelModel(
        font='Source Sans Pro,10,-1,5,50,0,0,0,0,0', foreground='#000000',
        height=40.0, parent_component='DisplayComponent', text='UInt32 (RO)',
        width=88.0, x=20.0, y=170.0)
    scene403 = LabelModel(
        font='Source Sans Pro,10,-1,5,50,0,0,0,0,0', foreground='#000000',
        height=41.0, parent_component='DisplayComponent', text='Int64 (RO)',
        width=88.0, x=20.0, y=210.0)
    scene404 = LabelModel(
        font='Source Sans Pro,10,-1,5,50,0,0,0,0,0', foreground='#000000',
        height=40.0, parent_component='DisplayComponent', text='Float (RO)',
        width=88.0, x=20.0, y=251.0)
    scene405 = LabelModel(
        font='Source Sans Pro,10,-1,5,50,0,0,0,0,0', foreground='#000000',
        height=40.0, parent_component='DisplayComponent', text='Double (RO)',
        width=88.0, x=20.0, y=291.0)
    scene40 = BoxLayoutModel(
        direction=2, height=241.0, width=88.0, x=20.0, y=90.0,
        children=[scene400, scene401, scene402, scene403, scene404, scene405])
    scene410 = CheckBoxModel(
        height=40.0, keys=[f'{deviceId}.boolPropertyReadOnly'],
        parent_component='DisplayComponent', width=158.0, x=108.0, y=90.0)
    scene411 = DisplayLabelModel(
        font_size=10, height=40.0, keys=[f'{deviceId}.int32PropertyReadOnly'],
        parent_component='DisplayComponent', width=158.0, x=108.0, y=130.0)
    scene412 = DisplayLabelModel(
        font_size=10, height=40.0, keys=[f'{deviceId}.uint32PropertyReadOnly'],
        parent_component='DisplayComponent', width=158.0, x=108.0, y=170.0)
    scene413 = DisplayLabelModel(
        font_size=10, height=41.0, keys=[f'{deviceId}.int64PropertyReadOnly'],
        parent_component='DisplayComponent', width=158.0, x=108.0, y=210.0)
    scene414 = DisplayLabelModel(
        font_size=10, height=40.0, keys=[f'{deviceId}.floatPropertyReadOnly'],
        parent_component='DisplayComponent', width=158.0, x=108.0, y=251.0)
    scene415 = DisplayLabelModel(
        font_size=10, height=40.0, keys=[f'{deviceId}.doublePropertyReadOnly'],
        parent_component='DisplayComponent', width=158.0, x=108.0, y=291.0)
    scene41 = BoxLayoutModel(
        direction=2, height=241.0, width=158.0, x=108.0, y=90.0,
        children=[scene410, scene411, scene412, scene413, scene414, scene415])
    scene420 = LabelModel(
        font='Source Sans Pro,10,-1,5,50,0,0,0,0,0', foreground='#000000',
        height=40.0, parent_component='DisplayComponent', text='Bool',
        width=99.0, x=266.0, y=90.0)
    scene421 = LabelModel(
        font='Source Sans Pro,10,-1,5,50,0,0,0,0,0', foreground='#000000',
        height=40.0, parent_component='DisplayComponent', text='Int32',
        width=99.0, x=266.0, y=130.0)
    scene422 = LabelModel(
        font='Source Sans Pro,10,-1,5,50,0,0,0,0,0', foreground='#000000',
        height=40.0, parent_component='DisplayComponent', text='UInt32',
        width=99.0, x=266.0, y=170.0)
    scene423 = LabelModel(
        font='Source Sans Pro,10,-1,5,50,0,0,0,0,0', foreground='#000000',
        height=41.0, parent_component='DisplayComponent', text='Int64',
        width=99.0, x=266.0, y=210.0)
    scene424 = LabelModel(
        font='Source Sans Pro,10,-1,5,50,0,0,0,0,0', foreground='#000000',
        height=40.0, parent_component='DisplayComponent',
        text='Float (Min / Max)', width=99.0, x=266.0, y=251.0)
    scene425 = LabelModel(
        font='Source Sans Pro,10,-1,5,50,0,0,0,0,0', foreground='#000000',
        height=40.0, parent_component='DisplayComponent', text='Double',
        width=99.0, x=266.0, y=291.0)
    scene42 = BoxLayoutModel(
        direction=2, height=241.0, width=99.0, x=266.0, y=90.0,
        children=[scene420, scene421, scene422, scene423, scene424, scene425])
    scene430 = CheckBoxModel(
        height=40.0, keys=[f'{deviceId}.boolProperty'],
        parent_component='DisplayComponent', width=158.0, x=365.0, y=90.0)
    scene431 = DisplayLabelModel(
        font_size=10, height=40.0, keys=[f'{deviceId}.int32Property'],
        parent_component='DisplayComponent', width=158.0, x=365.0, y=130.0)
    scene432 = DisplayLabelModel(
        font_size=10, height=40.0, keys=[f'{deviceId}.uint32Property'],
        parent_component='DisplayComponent', width=158.0, x=365.0, y=170.0)
    scene433 = DisplayLabelModel(
        font_size=10, height=41.0, keys=[f'{deviceId}.int64Property'],
        parent_component='DisplayComponent', width=158.0, x=365.0, y=210.0)
    scene434 = DisplayLabelModel(
        font_size=10, height=40.0, keys=[f'{deviceId}.floatProperty'],
        parent_component='DisplayComponent', width=158.0, x=365.0, y=251.0)
    scene435 = DisplayLabelModel(
        font_size=10, height=40.0, keys=[f'{deviceId}.doubleProperty'],
        parent_component='DisplayComponent', width=158.0, x=365.0, y=291.0)
    scene43 = BoxLayoutModel(
        direction=2, height=241.0, width=158.0, x=365.0, y=90.0,
        children=[scene430, scene431, scene432, scene433, scene434, scene435])
    scene440 = CheckBoxModel(
        height=40.0, keys=[f'{deviceId}.boolProperty'],
        klass='EditableCheckBox',
        parent_component='EditableApplyLaterComponent', width=158.0, x=523.0,
        y=90.0)
    scene441 = IntLineEditModel(
        height=40.0, keys=[f'{deviceId}.int32Property'],
        parent_component='EditableApplyLaterComponent', width=158.0, x=523.0,
        y=130.0)
    scene442 = IntLineEditModel(
        height=40.0, keys=[f'{deviceId}.uint32Property'],
        parent_component='EditableApplyLaterComponent', width=158.0, x=523.0,
        y=170.0)
    scene443 = IntLineEditModel(
        height=41.0, keys=[f'{deviceId}.int64Property'],
        parent_component='EditableApplyLaterComponent', width=158.0, x=523.0,
        y=210.0)
    scene444 = DoubleLineEditModel(
        height=40.0, keys=[f'{deviceId}.floatProperty'],
        parent_component='EditableApplyLaterComponent', width=158.0, x=523.0,
        y=251.0)
    scene445 = DoubleLineEditModel(
        height=40.0, keys=[f'{deviceId}.doubleProperty'],
        parent_component='EditableApplyLaterComponent', width=158.0, x=523.0,
        y=291.0)
    scene44 = BoxLayoutModel(
        direction=2, height=241.0, width=158.0, x=523.0, y=90.0,
        children=[scene440, scene441, scene442, scene443, scene444, scene445])
    scene4 = BoxLayoutModel(
        height=241.0, width=661.0, x=20.0, y=90.0,
        children=[scene40, scene41, scene42, scene43, scene44])
    scene500 = LabelModel(
        font='Source Sans Pro,10,-1,5,50,0,0,0,0,0', foreground='#000000',
        height=40.0, parent_component='DisplayComponent', text='String',
        width=114.0, x=10.0, y=360.0)
    scene501 = LabelModel(
        font='Source Sans Pro,10,-1,5,50,0,0,0,0,0', foreground='#000000',
        height=41.0, parent_component='DisplayComponent', text='Regex String',
        width=114.0, x=10.0, y=400.0)
    scene502 = LabelModel(
        font='Source Sans Pro,10,-1,5,50,0,0,0,0,0', foreground='#000000',
        height=40.0, parent_component='DisplayComponent',
        text='Vector Regex String', width=114.0, x=10.0, y=441.0)
    scene50 = BoxLayoutModel(
        direction=2, height=121.0, width=114.0, x=10.0, y=360.0,
        children=[scene500, scene501, scene502])
    scene510 = DisplayLabelModel(
        font_size=10, height=40.0, keys=[f'{deviceId}.stringProperty'],
        parent_component='DisplayComponent', width=274.0, x=124.0, y=360.0)
    scene511 = DisplayLabelModel(
        font_size=10, height=41.0, keys=[f'{deviceId}.regexProperty'],
        parent_component='DisplayComponent', width=274.0, x=124.0, y=400.0)
    scene512 = DisplayListModel(
        height=40.0, keys=[f'{deviceId}.vectorRegexProperty'],
        parent_component='DisplayComponent', width=274.0, x=124.0, y=441.0)
    scene51 = BoxLayoutModel(
        direction=2, height=121.0, width=274.0, x=124.0, y=360.0,
        children=[scene510, scene511, scene512])
    scene520 = LineEditModel(
        height=40.0, keys=[f'{deviceId}.stringProperty'],
        klass='EditableLineEdit',
        parent_component='EditableApplyLaterComponent', width=273.0, x=398.0,
        y=360.0)
    scene521 = EditableRegexModel(
        height=41.0, keys=[f'{deviceId}.regexProperty'],
        parent_component='EditableApplyLaterComponent', width=273.0, x=398.0,
        y=400.0)
    scene522 = EditableListModel(
        height=40.0, keys=[f'{deviceId}.vectorRegexProperty'],
        parent_component='EditableApplyLaterComponent', width=273.0, x=398.0,
        y=441.0)
    scene52 = BoxLayoutModel(
        direction=2, height=121.0, width=273.0, x=398.0, y=360.0,
        children=[scene520, scene521, scene522])
    scene5 = BoxLayoutModel(
        height=121.0, width=661.0, x=10.0, y=360.0,
        children=[scene50, scene51, scene52])
    scene60 = TableElementModel(
        height=194.0, keys=[f'{deviceId}.table'],
        parent_component='DisplayComponent', width=366.0, x=20.0, y=510.0)
    scene61 = TableElementModel(
        height=194.0, keys=[f'{deviceId}.table'], klass='EditableTableElement',
        parent_component='EditableApplyLaterComponent', width=365.0, x=386.0,
        y=510.0)
    scene6 = BoxLayoutModel(
        height=194.0, width=731.0, x=20.0, y=510.0,
        children=[scene60, scene61])
    scene7 = LineModel(
        stroke='#000000', x1=10.0, x2=760.0, y1=490.0, y2=490.0)
    scene8 = LineModel(
        stroke='#000000', x1=10.0, x2=760.0, y1=80.0, y2=80.0)
    scene9 = LineModel(
        stroke='#000000', x1=10.0, x2=770.0, y1=350.0, y2=350.0)
    scene10 = LineModel(
        stroke='#000000', x1=800.0, x2=800.0, y1=40.0, y2=690.0)
    scene11 = DoubleLineEditModel(
        height=27.0,
        keys=[f'{deviceId}.outputFrequency'],
        parent_component='EditableApplyLaterComponent',
        width=121.0, x=1050.0, y=90.0)
    scene12 = LabelModel(
        font='Source Sans Pro,12,-1,5,50,0,1,0,0,0,Normal', height=20.0,
        parent_component='DisplayComponent', text='Output Channel',
        width=220.0, x=820.0, y=40.0)
    scene13 = LineModel(
        stroke='#000000', x1=810.0, x2=1330.0, y1=80.0, y2=80.0)
    scene140 = DisplayLabelModel(
        font_size=10, height=27.0, keys=[f'{deviceId}.outputFrequency'],
        parent_component='DisplayComponent', width=99.0, x=942.0, y=90.0)
    scene141 = DisplayLabelModel(
        font_size=10, height=27.0, keys=[f'{deviceId}.outputCounter'],
        parent_component='DisplayComponent', width=99.0, x=942.0, y=117.0)
    scene142 = DisplayLabelModel(
        font_size=10, height=27.0, keys=[f'{deviceId}.currentInputId'],
        parent_component='DisplayComponent', width=99.0, x=942.0, y=144.0)
    scene143 = DisplayLabelModel(
        font_size=10, height=27.0, keys=[f'{deviceId}.inputCounter'],
        parent_component='DisplayComponent', width=99.0, x=942.0, y=171.0)
    scene144 = DisplayLabelModel(
        font_size=10, height=27.0, keys=[f'{deviceId}.inputCounterAtEos'],
        parent_component='DisplayComponent', width=99.0, x=942.0, y=198.0)
    scene14 = BoxLayoutModel(
        direction=2, height=135.0, width=99.0, x=942.0, y=90.0,
        children=[scene140, scene141, scene142, scene143, scene144])
    scene150 = LabelModel(
        font='Source Sans Pro,10,-1,5,50,0,0,0,0,0', foreground='#000000',
        height=27.0, parent_component='DisplayComponent',
        text='Output frequency', width=122.0, x=820.0, y=90.0)
    scene151 = LabelModel(
        font='Source Sans Pro,10,-1,5,50,0,0,0,0,0', foreground='#000000',
        height=27.0, parent_component='DisplayComponent',
        text='Output counter', width=122.0, x=820.0, y=117.0)
    scene152 = LabelModel(
        font='Source Sans Pro,10,-1,5,50,0,0,0,0,0', foreground='#000000',
        height=27.0, parent_component='DisplayComponent',
        text='Current Input Id', width=122.0, x=820.0, y=144.0)
    scene153 = LabelModel(
        font='Source Sans Pro,10,-1,5,50,0,0,0,0,0', foreground='#000000',
        height=27.0, parent_component='DisplayComponent', text='Input counter',
        width=122.0, x=820.0, y=171.0)
    scene154 = LabelModel(
        font='Source Sans Pro,10,-1,5,50,0,0,0,0,0', foreground='#000000',
        height=27.0, parent_component='DisplayComponent',
        text='Input counter @ EOS', width=122.0, x=820.0, y=198.0)
    scene15 = BoxLayoutModel(
        direction=2, height=135.0, width=122.0, x=820.0, y=90.0,
        children=[scene150, scene151, scene152, scene153, scene154])
    scene16 = NDArrayGraphModel(
        height=191.0, keys=[f'{deviceId}.output.schema.node.ndarray'],
        parent_component='DisplayComponent', width=521.0, x=820.0, y=560.0)
    scene17 = VectorGraphModel(
        height=191.0, keys=[f'{deviceId}.output.schema.node.vecInt64'],
        parent_component='DisplayComponent', width=531.0, x=810.0, y=340.0)
    scene18 = DisplayCommandModel(
        height=31.0, keys=[f'{deviceId}.startWritingOutput'],
        parent_component='DisplayComponent', width=111.0, x=1060.0, y=140.0)
    scene19 = DisplayCommandModel(
        height=34.0, keys=[f'{deviceId}.stopWritingOutput'],
        parent_component='DisplayComponent', width=111.0, x=1180.0, y=140.0)
    scene20 = DisplayCommandModel(
        height=31.0, keys=[f'{deviceId}.eosOutput'],
        parent_component='DisplayComponent', width=111.0, x=1060.0, y=180.0)
    scene21 = DisplayCommandModel(
        height=31.0, keys=[f'{deviceId}.writeOutput'],
        parent_component='DisplayComponent', width=111.0, x=1180.0, y=180.0)
    scene22 = DisplayCommandModel(
        height=31.0, keys=[f'{deviceId}.resetChannelCounters'],
        parent_component='DisplayComponent', width=111.0, x=1060.0, y=220.0)
    scene2300 = LabelModel(
        font='Source Sans Pro,10,-1,5,50,0,0,0,0,0', foreground='#000000',
        height=21.0, parent_component='DisplayComponent', text='int32',
        width=41.0, x=820.0, y=260.0)
    scene2301 = LabelModel(
        font='Source Sans Pro,10,-1,5,50,0,0,0,0,0', foreground='#000000',
        height=21.0, parent_component='DisplayComponent', text='string',
        width=41.0, x=820.0, y=281.0)
    scene230 = BoxLayoutModel(
        direction=2, height=42.0, width=41.0, x=820.0, y=260.0,
        children=[scene2300, scene2301])
    scene2310 = DisplayLabelModel(
        font_size=10, height=21.0,
        keys=[f'{deviceId}.output.schema.node.int32'],
        parent_component='DisplayComponent', width=170.0, x=861.0, y=260.0)
    scene2311 = DisplayLabelModel(
        font_size=10, height=21.0,
        keys=[f'{deviceId}.output.schema.node.string'],
        parent_component='DisplayComponent', width=170.0, x=861.0, y=281.0)
    scene231 = BoxLayoutModel(
        direction=2, height=42.0, width=170.0, x=861.0, y=260.0,
        children=[scene2310, scene2311])
    scene23 = BoxLayoutModel(
        height=42.0, width=211.0, x=820.0, y=260.0,
        children=[scene230, scene231])
    scene24 = LabelModel(
        font='Source Sans Pro,10,-1,5,50,0,0,0,0,0,Normal', height=18.0,
        parent_component='DisplayComponent', text='NDArray', width=111.0,
        x=810.0, y=540.0)
    scene25 = LabelModel(
        font='Source Sans Pro,10,-1,5,50,0,0,0,0,0,Normal', height=18.0,
        parent_component='DisplayComponent', text='VectorInt64', width=111.0,
        x=810.0, y=310.0)
    scene = SceneModel(
        height=766.0, width=1355.0,
        children=[scene0, scene1, scene2, scene3, scene4, scene5, scene6,
                  scene7, scene8, scene9, scene10, scene11, scene12, scene13,
                  scene14, scene15, scene16, scene17, scene18, scene19,
                  scene20, scene21, scene22, scene23, scene24, scene25])
    return write_scene(scene)



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.
import uuid
from asyncio import gather, set_event_loop, sleep, wait_for
from contextlib import ExitStack, contextmanager
from functools import partial, wraps
from unittest import TestCase
from unittest.mock import Mock

from karabo.middlelayer.device_server import MiddleLayerDeviceServer
from karabo.middlelayer.eventloop import EventLoop
from karabo.middlelayer.utils import ensure_coroutine


def async_tst(f=None, *, timeout=None):
    if f is None:
        assert timeout is not None, 'timeout must be given!'
        return partial(async_tst, timeout=timeout)

    f = ensure_coroutine(f)
    return sync_tst(f, timeout=timeout)


def sync_tst(f=None, *, timeout=None):
    if f is None:
        assert timeout is not None, 'timeout must be given!'
        return partial(sync_tst, timeout=timeout)

    # Default timeout is 30 seconds. Slower tests can give longer values
    timeout = 30 if timeout is None else timeout

    @wraps(f)
    def wrapper(self):
        task = self.loop.create_task(
            self.loop.run_coroutine_or_thread(f, self), self.lead)
        self.loop.run_until_complete(wait_for(task, timeout))

    return wrapper


class DeviceTest(TestCase):
    timeout = 20

    @classmethod
    def setUpClass(cls):
        with ExitStack() as cls.exit_stack:
            cls.loop = EventLoop()
            set_event_loop(cls.loop)
            cls.exit_stack.enter_context(cls.lifetimeManager())
            cls.exit_stack = cls.exit_stack.pop_all()

    @classmethod
    def tearDownClass(cls):
        try:
            with cls.exit_stack:
                pass
        finally:
            cls.loop.close()
            set_event_loop(None)

    @classmethod
    @contextmanager
    def lifetimeManager(cls):
        """This context manager is run around the test class"""
        cls.lead = Mock()
        cls.lead.deviceId = f"test-mdl-{uuid.uuid4()}"
        cls.lead._sigslot = Mock()
        cls.lead._sigslot.loop = cls.loop
        cls.lead._sigslot.tasks = set()
        yield

    @classmethod
    @contextmanager
    def deviceManager(cls, *devices, lead):
        """Manage the devices to run during the tests

        `devices` are the devices that should be run during the tests,
        and `lead` is the device under which name the tests are running.
        """
        if lead not in devices:
            devices += (lead,)
        cls.loop.run_until_complete(
            gather(*(d.startInstance() for d in devices)))
        cls.lead = lead

        # Make sure that we definately release here with a total
        # sleep time. All times in seconds
        sleep_time = 0.1
        total_time = cls.timeout
        while total_time >= 0:
            onlines = [d.is_initialized for d in devices]
            if all(onlines):
                break
            total_time -= sleep_time
            cls.loop.run_until_complete(sleep(sleep_time))
        yield
        cls.loop.run_until_complete(
            gather(*(d.slotKillDevice() for d in devices)))
        del cls.lead


def setEventLoop():
    loop = EventLoop()
    set_event_loop(loop)
    return loop


def create_device_server(serverId, plugins=[], config={}):
    """Create a device server instance with `plugins`

    :param serverId: the serverId of the server
    :param config: The server configuration
    :param plugins: Optional plugins list of device classes
    """
    config.update({"serverId": serverId,
                   "heartbeatInterval": 20})
    server = MiddleLayerDeviceServer(config)
    if plugins:

        async def scanPluginsOnce(server):
            return False

        server.scanPluginsOnce = scanPluginsOnce.__get__(
            server, type(server))
        server.plugins = {klass.__name__: klass for klass in plugins}
    return server



from ..models import Scene
from ..utils import (
    datetime_from_str, datetime_now, datetime_to_str, get_scene_links)

SCENE_UUID = "73d31361-b019-4d0c-9fd5-aa0acd536bb8"
SCENE_NAME = "das_Bild"

SCENE_HEADER = (
    f'<xml item_type="scene" uuid="{SCENE_UUID}" '
    f'simple_name="{SCENE_NAME}">'
    '<svg:svg xmlns:svg="http://www.w3.org/2000/svg" '
    'xmlns:krb="http://karabo.eu/scene" height="768" width="1024" '
    'krb:uuid="73d31361-b019-4d0c-9fd5-aa0acd536bb8" krb:version="2">')

SCENE_FOOTER = '</svg:svg></xml>'


def test_date_tools_round_trip():
    """Test the datetime tools with a roundtrip"""
    dt = datetime_now()
    ts = datetime_to_str(dt)
    assert datetime_to_str(dt) == ts
    assert datetime_from_str(ts) == dt


def test_get_scene_links():
    """Test the extraction of UUIDs of target scenes of scene links
    in a scene"""

    # Invalid scene links must be ignored
    svg = (
        f'{SCENE_HEADER}'
        '<svg:rect krb:class="SceneLink" '
        f'krb:target="scene_tgt:091:" />'  # Missing the UUID
        '<svg:rect krb:class="SceneLink" '
        f'krb:target="{SCENE_UUID}" />'  # Missing the target name
        f'{SCENE_FOOTER}')
    links = get_scene_links(
        Scene(uuid=SCENE_UUID, name=f"{SCENE_NAME}", svg_data=svg))
    assert len(links) == 0, f"Expected to find no scene link, got {len(links)}"

    svg = (
        f'{SCENE_HEADER}'
        '<svg:rect krb:class="Scene__Link" '  # Wrong  krb:class
        f'krb:target="scene_tgt:{SCENE_UUID}" />'
        f'{SCENE_FOOTER}')
    links = get_scene_links(
        Scene(uuid=SCENE_UUID, name=f"{SCENE_NAME}", svg_data=svg))
    assert len(links) == 0, f"Expected to find no scene link, got {len(links)}"

    # Multiple scene links must be returned
    svg = (
        f'{SCENE_HEADER}'
        '<svg:rect krb:class="SceneLink" '
        f'krb:target="scene_tgt:{SCENE_UUID}" />'
        '<svg:rect krb:class="SceneLink" '
        f'krb:target="scene_tgt:{SCENE_UUID}" />'
        f'{SCENE_FOOTER}')
    links = get_scene_links(
        Scene(uuid=SCENE_UUID, name=f"{SCENE_NAME}", svg_data=svg))
    assert len(links) == 2, f"Expected to find 2 scene links, got {len(links)}"
    assert links[0] == SCENE_UUID
    assert links[1] == SCENE_UUID



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.
import base64
from datetime import UTC, datetime

from karabo.native import Hash, decodeBinary, encodeBinary

DATE_FORMAT = "%Y-%m-%d %H:%M:%S"


class ConfigurationDBError(Exception):
    pass


def hashToBase64Bin(hash):
    """Converts a hash to its binary serialized form encoded in Base64.

    :param hash: the hash to be converted.
    :returns: a string with the base64 encoding of the binary serialized form
        of the hash.
    """
    binHash = encodeBinary(hash)
    b64Bin = base64.b64encode(binHash).decode('utf-8')
    return b64Bin


def hashFromBase64Bin(encodedHash):
    """Retrieves a hash from its binary serialized form encoded in Base64.

    :param encodedHash: the encoded hash.
    :returns: the decoded hash.
    """
    buf = base64.b64decode(encodedHash)
    hash = decodeBinary(buf)
    return hash


def schemaToBase64Bin(sch):
    """Converts a schema to its binary serialized form encoded in Base64.

    :param sch: the schema to be converted.
    :returns: a string with the base64 encoding of the binary serialized form
        of the schema.
    """
    hashWithSch = Hash('sch', sch)
    binSch = encodeBinary(hashWithSch)
    b64Bin = base64.b64encode(binSch).decode('utf-8')
    return b64Bin


def schemaFromBase64Bin(encodedSch):
    """Retrieves a schema from its binary serialized form encoded in Base64.

    :param encodedSch: the encoded schema.
    :returns: the decoded schema.
    """
    buf = base64.b64decode(encodedSch)
    hashWithSch = decodeBinary(buf)
    sch = hashWithSch['sch']
    return sch


def datetime_now() -> datetime:
    """Return a datetime object in utc without microseconds"""
    return datetime.now(UTC).replace(microsecond=0)


def datetime_str_now() -> str:
    """Return a datetime string for now for config db format"""
    return datetime_to_str(datetime.now(UTC))


def datetime_from_str(timestamp: str) -> datetime:
    """Parses a datetime string in date formats."""
    return datetime.strptime(timestamp, DATE_FORMAT).replace(tzinfo=UTC)


def datetime_to_str(timestamp: datetime | None) -> str:
    """Format a datetime object to a local str representation"""
    if timestamp is None:
        return ""
    return timestamp.replace(tzinfo=UTC).strftime(DATE_FORMAT)



import pytest
from lxml import etree

from karabo.project_db.util import make_str_if_needed, make_xml_if_needed


def test_make_xml_if_needed():
    xml_rep = "<test>foo</test>"
    ret = make_xml_if_needed(xml_rep)
    assert ret.tag == "test"
    assert ret.text == "foo"


def test_malformed_xml_inputs():
    xml_rep = "<test>foo</test><bad|symbols></bad|symbols>"
    with pytest.raises(ValueError):
        make_xml_if_needed(xml_rep)


def test_make_str_if_needed():
    element = etree.Element('test')
    element.text = 'foo'
    str_rep = make_str_if_needed(element)
    assert str_rep == "<test>foo</test>\n"



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.
"""Starting and stopping Karabo
============================

Karabo uses daemontools to supervise its device servers. Every device
server corresponds to one directory in `$KARABO/var/service`.
Typically, the device server's ID is used as a name for this
directory, just slashes replaced by underscores.

Within this directory, there is one `run` file, which is executed by
supervise. Typically, they are just symbolic links to files in
`$KARABO/service`, which contains ready-made code to execute Karabo
devices servers, but you are free to put whatever code you want.  If
you do use the ready-mades, a `parameters` file simply contains the
command line parameters to the server to be started.

There should be a `log` directory beside the `run` file, which should
again contain a `run` file which is executed for logging the standard
output from the device server. Again, there is a ready-made code in
`$KARABO/service/logger`, which you may link to.

Given that said set-up is actually pretty complicated, the
`karabo-add-deviceserver` script creates it for you. You just pass the
name of the device server, its type (middlelayerserver, cppserver or
pythonserver) and additional command line parameters, and all the
files and directories mentioned above are created.

Those device servers are then started with `karabo-start`. This is
just a small wrapper around daemontools' `svc` command, feel free to
use the underlying command as you may please. `karabo-stop` stops the
servers again, and `karabo-check` tells you which servers are
currently running and how they are doing.

If the standard logger described above is used, you may see the output
using `karabo-xterm`, which opens an xterm per started device server
to continuously show its output, or `karabo-gterm` which shows the
same as tabs in a gnome terminal.
"""

import os
import os.path as osp
import re
import shutil
import subprocess
import sys
from functools import wraps
from tempfile import mkdtemp
from textwrap import dedent
from time import sleep

COLORS = {
    "NORMAL": "\033[m",
    "RED": "\033[1;31m",
    "YELLOW": "\033[33m",
    "GREEN": "\033[0;32m",
}


def absolute(*path):
    return osp.join(os.environ["KARABO"], *path)


def entrypoint(func):
    @wraps(func)
    def invocation():
        if "KARABO" not in os.environ:
            print('Please activate Karabo '
                  '("source" the activate file in your Karabo directory)')
            return 1

        try:
            assert len(sys.argv) <= 1 or sys.argv[1] not in {"-h", "--help"}
            return func()
        except AssertionError:
            print(dedent("    " + func.__doc__))
            return 2
    return invocation


def supervise():
    if not check_service_dir():
        return False
    svok = subprocess.call([absolute("extern", "bin", "svok"),
                            absolute("var", "service", ".svscan")])
    if svok != 0:
        print("starting supervisor")
        supervise = subprocess.Popen(
            [absolute("extern", "bin", "supervise"),
             absolute("var", "service", ".svscan")],
            stdout=subprocess.PIPE)
        subprocess.Popen([absolute("extern", "bin", "multilog"),
                          absolute("var", "log", "svscan")],
                         stdin=supervise.stdout.fileno())
        sleep(1)  # give it some time to actually start
    return True


def defaultall():
    if not check_service_dir():
        return []
    existing = available_services()
    argv = [arg.replace("/", "_") for arg in sys.argv[1:]
            if not arg.startswith("-")]
    argv = [arg if arg in existing else arg.lower()
            for arg in argv]
    if not argv:
        return sorted(existing)
    else:
        return argv


def exec_defaultall(cmd, *args):
    if not supervise():
        return
    path = absolute("extern", "bin", cmd)
    os.execv(path, [cmd] + list(args) + defaultall())


def isexecutable(fn):
    return os.path.isfile(fn) and os.access(fn, os.X_OK)


def check_service_dir():
    service_dir = absolute("var", "service")
    if osp.exists(service_dir) and osp.isdir(service_dir):
        return True
    else:
        print("service directory not present."
              " Please create one with `karabo-create-services`")
        return False


def available_services():
    """Get all available services"""
    os.chdir(absolute("var", "service"))
    existing = {d for d in os.listdir() if not d.startswith(".")}
    return existing


@entrypoint
def make_service_dir():
    """karabo-create-services - creates Karabo servers folder

      karabo-create-services [-h|--help] template*

    creates a Karabo servers folder for a specific template.
    If none is given, creates a default template.

    Options are:
        default - A standard backbone installation. NB: this
                  configuration is likely targeted to the
                  infrastructure of the European XFEL
        empty   - An empty service folder.
                  Services can be added with the
                  `karabo-add-deviceserver` option
        jms_local
                - A stand-alone installation.
                  it will start the JMS broker on the local port
                  7777. To use this configuration, set the
                  KARABO_BROKER file in the var/environment folder
                  to 'tcp://localhost:7777
    """
    if osp.isdir(absolute("var", "service")):
        print("service directory already existing")
        return

    template = "default"
    if len(sys.argv[1:]) == 1:
        template = sys.argv[1]

    if not osp.isdir(absolute("service.in", template)):
        print(f"template service '{template}' directory not recognised.")
        dirs = [dir
                for dir in os.listdir(absolute("service.in"))
                if osp.isdir(absolute("service.in", dir))]
        print(f"Available templates are : {' ,'.join(dirs)}")
        return

    src = absolute("service.in", template)
    target = absolute("var", "service")
    shutil.copytree(src, target)


@entrypoint
def startkarabo():
    """karabo-start - start Karabo device servers

      karabo-start [-h|--help] device-servers*

    start given device servers. If none is given, start all.

    If the device server is already running, nothing happens.
    """
    exec_defaultall("svc", "-u")


@entrypoint
def stopkarabo():
    """karabo-stop - stop Karabo device servers

      karabo-stop [-h|--help] [-a|--all] [deviceserver1 [deviceserver2 [...]]]

    stops the given device servers.
    If the -a or --all flag is given, all device servers are stopped.
    Nothing happens for an already stopped device server.
    """
    assert len(sys.argv) > 1
    exec_defaultall("svc", "-d")


@entrypoint
def killkarabo():
    """karabo-kill - kill Karabo device servers

      karabo-kill [-h|--help] [-udopcaitkx ]

    send a signal to the listed device servers, by default to all.

    The signals are as follows:

    -i
      send an INT signal (as if Ctrl-C had been pressed)

    -t
      send a TERM signal (this should stop a server cleanly)

    -k
      send a KILL signal (kill the server for good)

    -p
      send a STOP signal

    -c
      send a CONT signal

    -a
      send an ALRM signal

    Several signals can be given on the command line. A "-l" signifies that
    not the device server, but its logger is meant. A + sign before a signal
    means (as in -+p) that the signal should be sent to the entire process
    group. This is implicit for the "-k" option to always kill all bound
    Python devices of a device server.

    There are other commands that can be given to a device server:

    -u
      Up. This is what karabo-start does.

    -d
      Down. This is what karabo-stop does.

    -o
      Start the device server once, but don't restart it if it stops.

    -x
      Kill the supervisor of the device server. Note that it will be restarted
      by svscan.

    If you want to kill all of karabo, use the following:

      karabo-kill -dx .svscan  # prevent supervisors from getting restarted
      karabo-kill -dx          # kill all device servers and their supervisors
      karabo-kill -l -dx       # kill all loggers and their supervisors

    This command does not try to start supervisors before running, so it is
    the only command usable to shut down Karabo completely.
    """
    assert len(sys.argv) > 1
    path = absolute("extern", "bin", "svc")
    args = [arg for arg in sys.argv if arg.startswith("-")]
    dirs = defaultall()
    if "-l" in args:
        args.remove("-l")
        dirs = [osp.join(d, "log") for d in dirs]

    # Add '+' option to '-k' arguments
    re1 = re.compile('-k+$')
    args = ['-+k' if re1.match(a) else a for a in args]

    # Split arguments containing 'k' and add '+' option to the latter
    re2 = re.compile('-[a-z]*k[a-z]*')
    for i in range(len(args)):
        if re2.match(args[i]) is not None:
            args[i] = args[i].replace('k', '')
            args.append('-+k')
    args = list(set(args))  # get rid of duplicates

    os.execv(path, ["svc"] + args + dirs)


def colorize(status_line):
    """
    Colorize the line of the following structure:
    `karabo_server: up (pid 26862) 448450 seconds, normally down, running`
    according to the up/down status
    """
    def color_string(color_name, line):
        return COLORS[color_name] + line + COLORS["NORMAL"]

    match = re.search(r"^(.+:)(\s+(\w+)\s+((?:\(.+\)\s)*)(\d+).+\s(.+))$",
                      status_line)
    if not match:
        return status_line
    server = match.group(1)
    rest = match.group(2)
    status = match.group(3)
    uptime = int(match.group(5))
    service_status = match.group(6)
    if status == "up" and service_status == "stopping":
        server = color_string("YELLOW", server)
    elif status == "up" and uptime > 1:
        server = color_string("GREEN", server)
    elif status == "up":
        server = color_string("YELLOW", server)
    elif status == "down":
        server = color_string("RED", server)
    else:
        server = COLORS["RED"] + server + COLORS["NORMAL"]
    return f"{server}{rest}"


@entrypoint
def checkkarabo():
    """karabo-check - check Karabo device servers

      karabo-check [-h|--help] device-servers*
    this shows the status of given Karabo device servers.

    If no device server is given, show the status of all device servers.
    """
    if not supervise():
        return
    svstat = subprocess.run(
        [absolute("extern", "bin", "svstat")] + defaultall(),
        stdout=subprocess.PIPE, encoding="utf8")
    print('\n'.join(map(colorize, svstat.stdout.strip().split('\n'))))


@entrypoint
def gnometermlog():
    """karabo-gterm - show logs in tabs of a gnome terminal

      karabo-gterm [-h|--help] device-server*

    This shows the logs of the given Karabo device servers in the
    tabs of a gnome terminal. If no device server is given, the
    logs of all device servers are shown.
    All this only works for device servers where logging is
    enabled.
    """
    if shutil.which("gnome-terminal") is None:
        print("'gnome-terminal' cannot be found - abort!")
        print("Try to use 'karabo-xterm' instead.")
        return 1
    use_lnav = False
    if shutil.which("lnav") is not None:
        use_lnav = True
    servers = [fn for fn in defaultall()
               if isexecutable(osp.join(fn, "log", "run"))]
    if use_lnav:
        cmd = sum((["--tab", "-e",
                    r'bash -c "echo -en \\\e]0\;{0}\\\a;'  # set tab title
                    'lnav $KARABO/var/log/{0}/current"'
                    .format(s)]
                   for s in servers), ["gnome-terminal"])
    else:
        cmd = sum((["--tab", "-e",
                    r'bash -c "echo -en \\\e]0\;{0}\\\a;'  # set tab title
                    'tail -n 100 -F $KARABO/var/log/{0}/current"'
                    .format(s)]
                   for s in servers), ["gnome-terminal"])
    os.execvp("gnome-terminal", cmd)


@entrypoint
def xtermlog():
    """karabo-xterm - show logs in an xterm

      karabo-xterm [-h|--help] device-server*

    This shows the logs of all given device servers in a xterm.
    This only works if logging is enabled for that device server.
    If no device server is given, the logs of all device servers
    are shown in their xterms.
    """
    if shutil.which("xterm") is None:
        print("'xterm' cannot be found - abort!")
        return 1
    use_lnav = False
    if shutil.which("lnav") is not None:
        use_lnav = True
    servers = [fn for fn in defaultall()
               if isexecutable(osp.join(fn, "log", "run"))]
    for server in servers:
        if use_lnav:
            subprocess.Popen(["xterm", "-T", server, "-e", "lnav",
                              absolute("var", "log", server, "current")])
        else:
            # Capital '-F' to follow renaming - 'current' is a rolling log file
            subprocess.Popen(["xterm", "-T", server, "-e", "tail", "-n", "100",
                              "-F", absolute("var", "log", server, "current")])


@entrypoint
def less():
    """karabo-less - show log in the current terminal

      karabo-less [-h|--help] device-server

    This shows the log of the given device server in the current terminal.
    This only works if logging is enabled for that device server.

    "lnav" will be used to display the log, "less --follow-name +F" otherwise.
    In the latter case, use "CTRL-C" to un-follow, "F" to follow again,
    "q" to quit.
    """
    servers = [fn for fn in defaultall()
               if isexecutable(osp.join(fn, "log", "run"))]

    if len(servers) != 1:
        print("Please pass a single valid device server to the command.")
        return

    fn = absolute("var", "log", servers[0], "current")
    if shutil.which("lnav") is not None:
        subprocess.call(["lnav", fn])
    else:
        cmd = ("less", "--follow-name", "+F", fn)  # follow file
        while True:
            try:
                subprocess.call(cmd)
                break  # "less" was quit normally
            except KeyboardInterrupt:
                # The user hit "CTRL-C" to unfollow the log. "less" will be
                # restarted without "+F" to allow qutting the loop.
                cmd = ("less", "--follow-name", "+G", fn)
        subprocess.call(["reset"])


server_template = """#!/bin/bash
# this file has been generated by karabo-add-deviceserver
cd $KARABO/var/data
exec envdir $KARABO/var/environment karabo-{server_type} serverId={server_id}\
 {options} 2>&1
"""

logger_template = """#!/bin/bash
# this file has been generated by karabo-add-deviceserver
exec multilog $KARABO/var/log/{target_dir}
"""


@entrypoint
def adddeviceserver():
    """karabo-add-deviceserver - create a new Karabo device server

      karabo-add-deviceserver [-h|--help]
      karabo-add-deviceserver name type arguments*

    This creates a new device server in $KARABO/var/service.

    name is the name of the new device server
    type is one of: cppserver, middlelayerserver, macroserver,
    pythonserver, webserver or webaggregatorserver.
    arguments are passed to said device server

    The serverId will be set to name.
    """
    assert len(sys.argv) > 2

    _, server_id, server_type, *options = sys.argv

    if not check_service_dir():
        return 3

    assert server_type in {
        "cppserver", "macroserver", "middlelayerserver", "pythonserver",
        "webaggregatorserver", "webserver"}

    target_dir = server_id.lower().replace("/", "_")

    abs_target = absolute("var", "service", target_dir)

    if osp.exists(abs_target):
        print(f"ERROR service/{target_dir} already exists")
        return 3
    # hide the new dir from svscan by prefixing with a . until we are done
    tmpdir = mkdtemp(dir=absolute("var", "service"), prefix=".tmp-")

    try:
        with open(osp.join(tmpdir, "run"), "w") as fout:
            fout.write(server_template.format(
                server_type=server_type, server_id=server_id,
                options="'{}'".format("' '".join(options)) if options else ""))
            # make file executable, but not writable
            os.chmod(fout.fileno(), 0o555)
        open(osp.join(tmpdir, "down"), "w").close()
        open(osp.join(tmpdir, "orphanage"), "w").close()
        with open(osp.join(tmpdir, "name"), "w") as fout:
            fout.write(server_id)
        os.mkdir(osp.join(tmpdir, "log"))
        with open(osp.join(tmpdir, "log", "run"), "w") as fout:
            fout.write(logger_template.format(target_dir=target_dir))
            os.chmod(fout.fileno(), 0o555)
        os.rename(tmpdir, abs_target)
    except Exception:
        shutil.rmtree(tmpdir)
        raise


@entrypoint
def removedeviceserver():
    """karabo-remove-deviceserver - remove a Karabo device server

      karabo-remove-deviceserver [-h|--help] name

    This removes a device server from $KARABO/var/service.

    name is the name of the device server to remove.
    """
    assert len(sys.argv) == 2
    if not check_service_dir():
        return 3

    name = sys.argv[1].replace("/", "_")
    # hide the service from svscan by prefixing with a . so it doesn't restart
    tmppath = absolute("var", "service", f".{name}")
    if osp.exists(absolute("var", "service", name)):
        os.rename(absolute("var", "service", name), tmppath)
    else:
        name = name.lower()
        os.rename(absolute("var", "service", name), tmppath)
    subprocess.call([absolute("extern", "bin", "svc"), "-dx", tmppath])
    subprocess.call([absolute("extern", "bin", "svc"), "-dx",
                     osp.join(tmppath, "log")])
    shutil.rmtree(tmppath)



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.
from karabo.common.sanity_check import validate_macro
from karabo.middlelayer import AccessMode, Device, Macro, String, isSet


class MetaMacro(Device):
    """This is the device that starts macros.

    It doesn't actually get instantiated itself, but when it is started will
    execute the code it is given and start the macros therein."""
    project = String(
        displayedName="Project Name",
        description="The project name of the macro it lives in.",
        accessMode=AccessMode.INITONLY)

    code = String(
        displayedName="Python Code",
        description="The code to be executed. It typically defines Macros",
        accessMode=AccessMode.INITONLY)

    module = String(
        displayedName="Name of Module",
        description="The name of the module within the project",
        accessMode=AccessMode.INITONLY)

    uuid = String(
        displayedName="Project DB UUID",
        description="The UUID for this macro",
        accessMode=AccessMode.INITONLY)

    def __init__(self, config):
        super().__init__(config)
        Macro.klass = None
        not_valid = validate_macro(self.code)
        if not_valid:
            msg = ("The macro cannot be started.\n"
                   "{}").format("\n".join([issue for issue in not_valid]))
            raise TypeError(msg)

        try:
            code = compile(self.code, self.module, "exec")
            exec(code, {})
            # Get the macro top level class
            self.klass = Macro.klass
        finally:
            Macro.klass = None

    def startInstance(self, server=None, broadcast=True):
        # this does not call super, as we don't want to run MetaMacro itself,
        # but only the macros in the supplied code
        deviceId = self.deviceId.value
        klass = self.klass
        parameters = {
            "_deviceId_":  f"{deviceId}",
            "serverId": self.serverId.value,
            "hostName": self.hostName.value,
            "uuid": self.uuid.value,
            "module": self.module.value,
        }
        if isSet(self.project):
            parameters["project"] = self.project.value
        macro = klass(parameters)
        macro.store_code(self.code.value)
        return macro.startInstance(server, broadcast=broadcast)
